-------------------------------------------------------------------------------
--! @file       CipherCore.vhd
--! @brief      Cipher core for MORUS1280-128 
--! @author     Tao Huang 
--! @version    1.1 
--! @license    This project is released under the GNU Public License.
--!             The license and distribution terms for this file may be
--!             found in the file LICENSE in this distribution or at
--!             http://www.gnu.org/licenses/gpl-3.0.txt
--! @note       Modified based on the template written by Ekawat (ice) Homsirikamol
-------------------------------------------------------------------------------


library ieee;
use ieee.std_logic_1164.ALL;
use IEEE.NUMERIC_STD.ALL;
use work.AEAD_pkg.ALL;

entity CipherCore is
    generic (
        --! Reset behavior
        G_ASYNC_RSTN    : boolean := False; --! Async active low reset
        --! Block size (bits)
        G_DBLK_SIZE     : integer := 256;   --! Data
        G_KEY_SIZE      : integer := 128;   --! Key
        G_TAG_SIZE      : integer := 128;   --! Tag
        --! The number of bits required to hold block size expressed in
        --! bytes = log2_ceil(G_DBLK_SIZE/8)
        G_LBS_BYTES     : integer := 5; 
        --! Algorithm parameter to simulate run-time behavior
        --! Warning: Do not set any number higher than 32
        G_LAT_INIT      : integer := 16;    --! Pre-processing latency (init)
        G_LAT_FINAL     : integer := 8;     --! Post-processing latency (tag)
        --! Maximum supported AD/message/ciphertext length = 2^G_MAX_LEN-1
        G_MAX_LEN       : integer := SINGLE_PASS_MAX
    );
    port (
        --! Global
        clk             : in  std_logic;
        rst             : in  std_logic;
        --! PreProcessor (data)
        key             : in  std_logic_vector(G_KEY_SIZE       -1 downto 0);
        bdi             : in  std_logic_vector(G_DBLK_SIZE      -1 downto 0);
        --! PreProcessor (controls)
        key_ready       : out std_logic;
        key_valid       : in  std_logic;
        key_update      : in  std_logic;
        decrypt         : in  std_logic;
        bdi_ready       : out std_logic;
        bdi_valid       : in  std_logic;
        bdi_type        : in  std_logic_vector(3                -1 downto 0);
        bdi_partial     : in  std_logic;
        bdi_eot         : in  std_logic;
        bdi_eoi         : in  std_logic;
        bdi_size        : in  std_logic_vector(G_LBS_BYTES+1    -1 downto 0);
        bdi_valid_bytes : in  std_logic_vector(G_DBLK_SIZE/8    -1 downto 0);
        bdi_pad_loc     : in  std_logic_vector(G_DBLK_SIZE/8    -1 downto 0);
        --! PostProcessor
        bdo             : out std_logic_vector(G_DBLK_SIZE      -1 downto 0);
        bdo_valid       : out std_logic;
        bdo_ready       : in  std_logic;
        bdo_size        : out std_logic_vector(G_LBS_BYTES+1    -1 downto 0);
        msg_auth_done   : out std_logic;
        msg_auth_valid  : out std_logic
    );
end entity CipherCore;

architecture structure of CipherCore is
    --! Registers
    signal bdi_r        : std_logic_vector(G_DBLK_SIZE      -1 downto 0);
    signal key_r        : std_logic_vector(G_KEY_SIZE       -1 downto 0);
    signal tag          : std_logic_vector(G_DBLK_SIZE      -1 downto 0);
    signal tag_rev      : std_logic_vector(G_DBLK_SIZE      -1 downto 0);
    signal s0           : std_logic_vector(G_DBLK_SIZE      -1 downto 0);  --! state registers
    signal s1           : std_logic_vector(G_DBLK_SIZE      -1 downto 0);  --! state registers
    signal s2           : std_logic_vector(G_DBLK_SIZE      -1 downto 0);  --! state registers
    signal s3           : std_logic_vector(G_DBLK_SIZE      -1 downto 0);  --! state registers
    signal s4           : std_logic_vector(G_DBLK_SIZE      -1 downto 0);  --! state registers
    signal msg          : std_logic_vector(G_DBLK_SIZE      -1 downto 0);
    signal data         : std_logic_vector(G_DBLK_SIZE      -1 downto 0);
    signal len_ad       : std_logic_vector(61               -1 downto 0);
    signal len_d        : std_logic_vector(61               -1 downto 0);
    signal s0_r         : std_logic_vector(G_DBLK_SIZE      -1 downto 0);  
    signal s1_r         : std_logic_vector(G_DBLK_SIZE      -1 downto 0);  
    signal s2_r         : std_logic_vector(G_DBLK_SIZE      -1 downto 0);  
    signal s3_r         : std_logic_vector(G_DBLK_SIZE      -1 downto 0);  
    signal s4_r         : std_logic_vector(G_DBLK_SIZE      -1 downto 0);  
    signal msg_r        : std_logic_vector(G_DBLK_SIZE      -1 downto 0);  

    --! Signals
    signal to_output    : std_logic_vector(G_DBLK_SIZE      -1 downto 0);
    signal s0_out       : std_logic_vector(G_DBLK_SIZE      -1 downto 0);
    signal s1_out       : std_logic_vector(G_DBLK_SIZE      -1 downto 0);
    signal s2_out       : std_logic_vector(G_DBLK_SIZE      -1 downto 0);
    signal s3_out       : std_logic_vector(G_DBLK_SIZE      -1 downto 0);
    signal s4_out       : std_logic_vector(G_DBLK_SIZE      -1 downto 0);
    signal data_tmp     : std_logic_vector(255 downto 0); 

    --! Controls
    --!     Register
    signal is_decrypt   : std_logic;
    signal set_bdo_valid2_r : std_logic; 

    --!     Combinatorial
    signal ctr          : std_logic_vector(5        -1 downto 0);
    signal ld_key       : std_logic;
    signal ld_npub      : std_logic;
    signal en_block     : std_logic;
    signal ld_ctr       : std_logic;
    signal ld_msg       : std_logic; -- message will be reversed for correct endian
    signal ld_len       : std_logic; -- length info will not be reversed 
    signal en_ctr       : std_logic;
    signal en_tag       : std_logic;
    signal sel_decrypt  : std_logic;
    signal sel_final    : std_logic;
    signal bdi_ready_s  : std_logic; 
    signal update_s     : std_logic; 
    signal add_key      : std_logic;  
    signal add_s0       : std_logic; 
    signal add_key_c    : std_logic;  
    signal add_s0_c     : std_logic; 
    signal msg_zero     : std_logic; 
    signal msg_data     : std_logic; 
    signal msg_len      : std_logic; 
    signal msg_auth_done_s : std_logic; 
    signal set_bdo_valid1  : std_logic; 
    signal set_bdo_valid2  : std_logic; 
    
    type t_state is (S_INIT, S_WAIT_START, S_WAIT_KEY, 
        S_WAIT_NPUB, S_INIT_MSG, S_WAIT_MSG, S_FINAL,
        S_WAIT_TAG_AUTH);
    signal state        : t_state;
    signal nstate       : t_state;

    function reverse_byte(aa: std_logic_vector) return std_logic_vector is
            variable bb : std_logic_vector(aa'high downto aa'low);
    begin
        for i in ((aa'high+1)/8-1) downto 0 loop
            bb(i*8+7 downto i*8) := aa(aa'high-i*8 downto (aa'high+1)-8-i*8);
        end loop;
        return bb;
    end function reverse_byte;

begin
    --! =======================================================================
    --! Datapath
    --! =======================================================================
    process(clk)
    begin
        if rising_edge(clk) then

            if (ld_key = '1') then
                key_r <= reverse_byte(key);
            end if;

            if (ld_npub = '1') then
                s0_r  <= reverse_byte(bdi); 
                s1_r  <= key_r & key_r;  -- 128-bit key
                s2_r  <= x"ffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffff"; 
                s3_r  <= x"0000000000000000000000000000000000000000000000000000000000000000";
                s4_r  <= x"dd28b57342311120f12fc26d55183ddb6279e990593722150d08050302010100";
                msg_r <= x"0000000000000000000000000000000000000000000000000000000000000000"; 
                len_ad <= (others => '0');
                len_d  <= (others => '0');
                is_decrypt <= decrypt;
            end if;

            set_bdo_valid2_r <= set_bdo_valid2; 

            if (update_s = '1') then 
                s0_r  <= s0_out; 
                s1_r  <= s1_out; 
                s2_r  <= s2_out; 
                s3_r  <= s3_out; 
                s4_r  <= s4_out; 
                if (add_key_c = '1') then 
                    add_key <= '1'; 
                else 
                    add_key <= '0'; 
                end if; 

                if (add_s0_c = '1') then 
                    add_s0 <= '1';
                else 
                    add_s0 <= '0';
                end if; 
                if (msg_data = '1') then 
                    msg_r <= data; 
                elsif (msg_zero = '1') then 
                    msg_r <= (others => '0'); 
                elsif (msg_len = '1') then 
                    msg_r <= s3_out xor ((x"00000000000000000000000000000000") & (len_d & "000" & len_ad & "000")); 
                end if;
            end if; 

            if (bdi_ready_s = '1') then
                if (bdi_type(2 downto 1) = BDI_TYPE_ASS) then
                    len_ad <= std_logic_vector(unsigned(len_ad) + unsigned(bdi_size));
                elsif (bdi_type(2 downto 1) = BDI_TYPE_DAT) then
                    len_d <= std_logic_vector(unsigned(len_d) + unsigned(bdi_size));
                end if;
            end if;

            if (msg_auth_done_s = '1') then 
                if (tag_rev(255 downto 128) = bdi(255 downto 128)) then 
                    msg_auth_valid <= '1'; 
                else 
                    msg_auth_valid <= '0'; 
                end if; 
                msg_auth_done <= '1';

            end if; 

        end if;
    end process;

    aBlock: block
        signal vbits : std_logic_vector(G_DBLK_SIZE-1 downto 0);
        signal pad   : std_logic_vector(G_DBLK_SIZE-1 downto 0);
        signal ext   : std_logic_vector(G_DBLK_SIZE-1 downto 0);
    begin
        
        gVbits:
        for i in 0 to G_DBLK_SIZE/8-1 generate
            --! Get valid bits from valid bytes
            vbits(8*(31-i)+7 downto 8*(31-i)) <= (others => bdi_valid_bytes(i));           
        end generate;

        bdi_r <= reverse_byte(bdi); 
        to_output <= msg_r xor s0_r xor (s1_r(63 downto 0) & s1_r(255 downto 64)) xor (s2_r and s3_r) when is_decrypt = '0' else msg_r ; 
        data_tmp  <= bdi_r xor s0_out xor (s1_out(63 downto 0) & s1_out(255 downto 64)) xor (s2_out and s3_out); 
        data <= (data_tmp and vbits) when sel_decrypt = '1' else bdi_r; 
        
    end block;

    bdo <= tag_rev when (sel_final = '1') 
                   else reverse_byte(to_output); 

    bdo_valid <= set_bdo_valid1 or set_bdo_valid2_r;                

    tag <= s1_r xor s2_r xor s3_r xor s4_r;
    tag_rev <= reverse_byte(tag); 

    s0 <= s0_r; 
    s1 <= s1_r; 
    s2 <= s2_r; 
    s3 <= s3_r; 
    s4 <= s4_r; 
    msg <= msg_r; 
    --! =======================================================================
    --! Control
    --! =======================================================================
    gSyncRst:
    if (not G_ASYNC_RSTN) generate
        process(clk)
        begin
            if rising_edge(clk) then
                if (rst = '1') then
                    state <= S_INIT;
                else
                    state <= nstate;
                end if;
            end if;
        end process;
    end generate;
    gAsyncRstn:
    if (G_ASYNC_RSTN) generate
        process(clk, rst)
        begin
            if (rst = '0') then
                state <= S_INIT;
            elsif rising_edge(clk) then
                state <= nstate;
            end if;
        end process;
    end generate;

    process(clk)
    begin
        if rising_edge(clk) then
            if (ld_ctr = '1') then
                ctr <= (others => '0');
            elsif (en_ctr = '1') then
                ctr <= std_logic_vector(unsigned(ctr) + 1);
            end if;
        end if;
    end process;

    bdi_ready <= bdi_ready_s;

    process(
        state, key_valid, key_update, is_decrypt, ctr,
        bdi_valid, bdi_type, bdi_eot, bdi_eoi, bdi_size,
        bdo_ready)
    begin
        --! Internal
        nstate      <= state;
        ld_key      <= '0';
        ld_npub     <= '0';
        en_block    <= '0';
        ld_ctr      <= '0';
        ld_msg      <= '0';
        ld_len      <= '0'; 
        en_tag      <= '0';
        en_ctr      <= '0';
        sel_decrypt <= '0';
        sel_final   <= '0';
        set_bdo_valid1 <= '0'; 
        set_bdo_valid2 <= '0'; 

        update_s    <= '0'; 
        msg_zero    <= '0'; 
        msg_data    <= '0'; 
        msg_len     <= '0'; 
        add_key_c   <= '0'; 
        add_s0_c    <= '0'; 
        bdi_ready_s <= '0';

        --! External
        key_ready   <= '0';
        msg_auth_done_s <= '0';

        case state is
            when S_INIT =>
                --! After reset
                ld_ctr    <= '1';
                nstate <= S_WAIT_START;

            when S_WAIT_START =>
                --! Needs to check whether a new input is available first
                --! prior to checking key to ensure the correct operational
                --! step.
                if (bdi_valid = '1') then
                    if (key_update = '1') then
                        nstate <= S_WAIT_KEY;
                    else
                        nstate <= S_WAIT_NPUB;
                    end if;
                end if;

            when S_WAIT_KEY =>
                --! Wait for key
                if (key_valid = '1') then
                    key_ready <= '1';
                    ld_key    <= '1';
                    nstate    <= S_WAIT_NPUB;
                end if;

            when S_WAIT_NPUB =>
                bdi_ready_s <= '1';
                if (bdi_valid = '1') then
                    ld_npub <= '1';
                    nstate  <= S_INIT_MSG;
                end if;

            when S_INIT_MSG =>
                en_ctr <= '1';
                if (unsigned(ctr) = G_LAT_INIT-1) then
                    ld_ctr <= '1';
                    nstate <= S_WAIT_MSG;
                elsif (unsigned(ctr) = G_LAT_INIT-2) then 
                    update_s <= '1'; 
                    msg_zero <= '1'; 
                    add_key_c  <= '1'; 
                else 
                    update_s <= '1'; 
                    msg_zero <= '1'; 
                end if;

            when S_WAIT_MSG =>
                if (bdi_valid = '1') then
                    ld_msg  <= '1';
                    if (bdi_type = BDI_TYPE_ASS0) then
                        bdi_ready_s <= '1';
                        msg_data <= '1'; 
                        update_s <= '1';
                    elsif (bdi_type = BDI_TYPE_DAT0) then
                        bdi_ready_s <= '1';
                        msg_data <= '1'; 
                        update_s <= '1';
                        set_bdo_valid2 <= '1'; 
                        sel_decrypt <= is_decrypt; 
                    else
                        update_s <= '1'; 
                        add_s0_c <= '1'; 
                        msg_len   <= '1'; 
                        nstate <= S_FINAL; 
                    end if;
                end if;

            when S_FINAL => 
                if (unsigned(ctr) = G_LAT_FINAL) then
                    if (bdo_ready = '1' or is_decrypt = '1') then
                        sel_final <= '1';
                        ld_ctr    <= '1';
                        if (is_decrypt = '1') then
                            en_tag    <= '1';
                            nstate    <= S_WAIT_TAG_AUTH;
                        else
                            set_bdo_valid1 <= '1';
                            nstate    <= S_INIT;
                        end if;
                    end if;
                else
                    en_ctr <= '1';
                    update_s <= '1'; 
                end if;

            when S_WAIT_TAG_AUTH =>
                bdi_ready_s <= '1';
                if (bdi_valid = '1') then
                    ld_msg  <= '1';                    
                    msg_auth_done_s <= '1';
                    nstate        <= S_INIT;
                end if;
            when others => 
            
        end case;

    end process;

    -- update the morus state using step function 
    morus_state_updater : 
    entity work.morus_round(Behavioral) 
    port map ( s0      => s0, 
               s1      => s1, 
               s2      => s2, 
               s3      => s3, 
               s4      => s4, 
               msg     => msg, 
               key_r   => key_r, 
               s0_out  => s0_out, 
               s1_out  => s1_out, 
               s2_out  => s2_out, 
               s3_out  => s3_out, 
               s4_out  => s4_out,
               add_key => add_key,
               add_s0  => add_s0
               ); 

end structure;