-------------------------------------------------------------------------------
--! @file       morus_round.vhd
--! @brief      state update function for MORUS1280-128 
--! @author     Tao Huang 
--! @license    This project is released under the GNU Public License.
--!             The license and distribution terms for this file may be
--!             found in the file LICENSE in this distribution or at
--!             http://www.gnu.org/licenses/gpl-3.0.txt
-------------------------------------------------------------------------------


library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.NUMERIC_STD.ALL;

-- Uncomment the following library declaration if instantiating
-- any Xilinx leaf cells in this code.
--library UNISIM;
--use UNISIM.VComponents.all;

entity morus_round is
  generic (
  	G_DBLK_SIZE              : integer := 256  --! Data Block size (bits)
  	);

  Port (
  	s0                       : in  std_logic_vector(G_DBLK_SIZE             -1 downto 0);
  	s1                       : in  std_logic_vector(G_DBLK_SIZE             -1 downto 0);
  	s2                       : in  std_logic_vector(G_DBLK_SIZE             -1 downto 0);
  	s3                       : in  std_logic_vector(G_DBLK_SIZE             -1 downto 0);
  	s4                       : in  std_logic_vector(G_DBLK_SIZE             -1 downto 0);

  	msg                      : in  std_logic_vector(G_DBLK_SIZE             -1 downto 0);
    key_r                    : in  std_logic_vector(128                     -1 downto 0); -- 

  	s0_out                   : out std_logic_vector(G_DBLK_SIZE             -1 downto 0);
  	s1_out                   : out std_logic_vector(G_DBLK_SIZE             -1 downto 0);
  	s2_out                   : out std_logic_vector(G_DBLK_SIZE             -1 downto 0);
  	s3_out                   : out std_logic_vector(G_DBLK_SIZE             -1 downto 0);
  	s4_out                   : out std_logic_vector(G_DBLK_SIZE             -1 downto 0);

    -- signal 
    add_key                  : in  std_logic; 
    add_s0                   : in  std_logic 
  	);

end entity morus_round;

architecture Behavioral of morus_round is
  

  function rotl_small ( x : std_logic_vector(G_DBLK_SIZE-1 downto 0); n : integer) 
    return std_logic_vector is 
    variable y : std_logic_vector(G_DBLK_SIZE-1 downto 0); 
  begin  -- rot_xxx_yy (Reversed direction !!)
    y := x((255-n) downto 192) & x(255 downto (256-n)) & x(191-n downto 128) 
         & x(191 downto 192-n) & x(127-n downto 64) & x(127 downto 128-n) 
         & x(63-n downto 0) & x(63 downto 64-n); 
    return y; 
  end rotl_small;

begin
  
  morus_step_update: process (s0, s1, s2, s3, s4, msg, add_key, add_s0)
    variable t0, t1, t2, t3, t4  : std_logic_vector(G_DBLK_SIZE-1  downto 0); 
    variable s10, s13            : std_logic_vector(G_DBLK_SIZE-1  downto 0); 
    variable s21, s24            : std_logic_vector(G_DBLK_SIZE-1  downto 0); 
    variable s30, s32            : std_logic_vector(G_DBLK_SIZE-1  downto 0); 
    variable s41, s43            : std_logic_vector(G_DBLK_SIZE-1  downto 0); 
    variable s52, s54            : std_logic_vector(G_DBLK_SIZE-1  downto 0); 
    variable ts4                 : std_logic_vector(G_DBLK_SIZE-1  downto 0); 
  begin 
      if (add_s0 = '1') then 
          ts4 := s0 xor s4; 
      else 
          ts4 := s4; 
      end if; 
      -- round 1 
      t0  := s0 xor s3 xor (s1 and s2); 
      s10 := rotl_small(t0, 13); 
      --s13 := s3 rol 64; 
      s13 := ( s3(G_DBLK_SIZE-1 - 64 downto 0) & s3(G_DBLK_SIZE-1 downto G_DBLK_SIZE-64) ); 
    
      -- round 2
      t1  := s1 xor ts4 xor (s2 and s13) xor msg; 
      s21 := rotl_small(t1, 46); 
      --s24 := s4 rol 128; 
      s24 := ( ts4(G_DBLK_SIZE-1 - 128 downto 0) & ts4(G_DBLK_SIZE-1 downto G_DBLK_SIZE-128) ); 
    
      -- round 3
      t2  := s2 xor s10 xor (s13 and s24) xor msg; 
      s32 := rotl_small(t2, 38); 
      --s30 := s10 rol 192; 
      s30 := ( s10(G_DBLK_SIZE-1 - 192 downto 0) & s10(G_DBLK_SIZE-1 downto G_DBLK_SIZE-192) ); 
    
      -- round 4
      t3  := s13 xor s21 xor (s24 and s30) xor msg; 
      s43 := rotl_small(t3, 7); 
      --s41 := s21 rol 128; 
      s41 := ( s21(G_DBLK_SIZE-1 - 128 downto 0) & s21(G_DBLK_SIZE-1 downto G_DBLK_SIZE-128) ); 
    
      -- round 5
      t4  := s24 xor s32 xor (s30 and s41) xor msg; 
      s54 := rotl_small(t4, 4); 
      --s52 := s32 rol 64; 
      s52 := ( s32(G_DBLK_SIZE-1 - 64 downto 0) & s32(G_DBLK_SIZE-1 downto G_DBLK_SIZE-64) ); 
    
      s0_out <= s30; 
      s1_out <= s41; 
      if (add_key = '1') then 
        s1_out <= s41 xor (key_r & key_r); 
      else 
        s1_out <= s41; 
      end if; 
      s2_out <= s52; 
      s3_out <= s43; 
      s4_out <= s54; 
  end process;

end Behavioral;
